=== modified file 'NEWS'
--- NEWS	2006-05-23 10:57:02 +0000
+++ NEWS	2006-05-24 08:14:45 +0000
@@ -20,6 +20,9 @@
      for the PQM status window. (Robert Collins).
 
    * 'bzr selftest --benchmark' will run a new benchmarking selftest.
+     'bzr selftest --benchmark --lsprof-timed' will use lsprofile to generate
+     profile data for the individual profiled calls, allowing for fine
+     grained analysis of performance.
      (Robert Collins, Martin Pool).
 
    * 'bzr commit' shows a progress bar. This is useful for commits over sftp

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2006-05-21 17:56:17 +0000
+++ bzrlib/builtins.py	2006-05-24 08:04:06 +0000
@@ -1784,10 +1784,14 @@
                                  'throughout the test suite.',
                             type=get_transport_type),
                      Option('benchmark', help='run the bzr bencharks.'),
-                    ]
+                     Option('lsprof-timed',
+                            help='generate lsprof output for benchmarked'
+                                 ' sections of code.'),
+                     ]
 
-    def run(self, testspecs_list=None, verbose=False, one=False,
-            keep_output=False, transport=None, benchmark=None):
+    def run(self, testspecs_list=None, verbose=None, one=False,
+            keep_output=False, transport=None, benchmark=None,
+            lsprof_timed=None):
         import bzrlib.ui
         from bzrlib.tests import selftest
         import bzrlib.benchmarks as benchmarks
@@ -1807,14 +1811,19 @@
                 pattern = ".*"
             if benchmark:
                 test_suite_factory = benchmarks.test_suite
+                if verbose is None:
+                    verbose = True
             else:
                 test_suite_factory = None
+                if verbose is None:
+                    verbose = False
             result = selftest(verbose=verbose, 
                               pattern=pattern,
                               stop_on_failure=one, 
                               keep_output=keep_output,
                               transport=transport,
-                              test_suite_factory=test_suite_factory)
+                              test_suite_factory=test_suite_factory,
+                              lsprof_timed=lsprof_timed)
             if result:
                 bzrlib.trace.info('tests passed')
             else:

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2006-05-18 07:36:29 +0000
+++ bzrlib/tests/__init__.py	2006-05-24 08:14:45 +0000
@@ -47,6 +47,11 @@
 import bzrlib.inventory
 import bzrlib.iterablefile
 import bzrlib.lockdir
+try:
+    import bzrlib.lsprof
+except ImportError:
+    # lsprof not available
+    pass
 from bzrlib.merge import merge_inner
 import bzrlib.merge3
 import bzrlib.osutils
@@ -223,6 +228,9 @@
         self.extractBenchmarkTime(test)
         if self.showAll:
             self.stream.writeln('   OK %s' % self._testTimeString())
+            for bench_called, stats in test._benchcalls:
+                self.stream.writeln('LSProf output for %s(%s, %s)' % bench_called)
+                stats.pprint(file=self.stream)
         elif self.dots and self.pb is None:
             self.stream.write('~')
         elif self.dots:
@@ -375,6 +383,8 @@
 
     _log_file_name = None
     _log_contents = ''
+    # record lsprof data when performing benchmark calls.
+    _gather_lsprof_in_benchmarks = False
 
     def __init__(self, methodName='testMethod'):
         super(TestCase, self).__init__(methodName)
@@ -385,6 +395,7 @@
         self._cleanEnvironment()
         bzrlib.trace.disable_default_logging()
         self._startLogFile()
+        self._benchcalls = []
         self._benchtime = None
 
     def _ndiff_strings(self, a, b):
@@ -541,12 +552,24 @@
         unittest.TestCase.tearDown(self)
 
     def time(self, callable, *args, **kwargs):
-        """Run callable and accrue the time it takes to the benchmark time."""
+        """Run callable and accrue the time it takes to the benchmark time.
+        
+        If lsprofiling is enabled (i.e. by --lsprof-time to bzr selftest) then
+        this will cause lsprofile statistics to be gathered and stored in
+        self._benchcalls.
+        """
         if self._benchtime is None:
             self._benchtime = 0
         start = time.time()
         try:
-            callable(*args, **kwargs)
+            if not self._gather_lsprof_in_benchmarks:
+                return callable(*args, **kwargs)
+            else:
+                # record this benchmark
+                ret, stats = bzrlib.lsprof.profile(callable, *args, **kwargs)
+                stats.sort()
+                self._benchcalls.append(((callable, args, kwargs), stats))
+                return ret
         finally:
             self._benchtime += time.time() - start
 
@@ -1031,8 +1054,9 @@
 
 def run_suite(suite, name='test', verbose=False, pattern=".*",
               stop_on_failure=False, keep_output=False,
-              transport=None):
+              transport=None, lsprof_timed=None):
     TestCaseInTempDir._TEST_NAME = name
+    TestCase._gather_lsprof_in_benchmarks = lsprof_timed
     if verbose:
         verbosity = 2
         pb = None
@@ -1054,7 +1078,8 @@
 def selftest(verbose=False, pattern=".*", stop_on_failure=True,
              keep_output=False,
              transport=None,
-             test_suite_factory=None):
+             test_suite_factory=None,
+             lsprof_timed=None):
     """Run the whole test suite under the enhanced runner"""
     global default_transport
     if transport is None:
@@ -1068,7 +1093,8 @@
             suite = test_suite_factory()
         return run_suite(suite, 'testbzr', verbose=verbose, pattern=pattern,
                      stop_on_failure=stop_on_failure, keep_output=keep_output,
-                     transport=transport)
+                     transport=transport,
+                     lsprof_timed=lsprof_timed)
     finally:
         default_transport = old_transport
 

=== modified file 'bzrlib/tests/test_selftest.py'
--- bzrlib/tests/test_selftest.py	2006-05-18 08:41:56 +0000
+++ bzrlib/tests/test_selftest.py	2006-05-24 08:04:06 +0000
@@ -395,7 +395,7 @@
         self.calls.append(('clear',))
 
 
-class TestResult(TestCase):
+class TestTestResult(TestCase):
 
     def test_progress_bar_style_quiet(self):
         # test using a progress bar.
@@ -468,6 +468,53 @@
         # cheat. Yes, wash thy mouth out with soap.
         self._benchtime = None
 
+    def _time_hello_world_encoding(self):
+        """Profile two sleep calls
+        
+        This is used to exercise the test framework.
+        """
+        self.time(unicode, 'hello', errors='replace')
+        self.time(unicode, 'world', errors='replace')
+
+    def test_lsprofiling(self):
+        """Verbose test result prints lsprof statistics from test cases."""
+        try:
+            import bzrlib.lsprof
+        except ImportError:
+            raise TestSkipped("lsprof not installed.")
+        result_stream = StringIO()
+        result = bzrlib.tests._MyResult(
+            unittest._WritelnDecorator(result_stream),
+            descriptions=0,
+            verbosity=2,
+            )
+        # we want profile a call of some sort and check it is output by
+        # addSuccess. We dont care about addError or addFailure as they
+        # are not that interesting for performance tuning.
+        # make a new test instance that when run will generate a profile
+        example_test_case = TestTestResult("_time_hello_world_encoding")
+        example_test_case._gather_lsprof_in_benchmarks = True
+        # execute the test, which should succeed and record profiles
+        example_test_case.run(result)
+        # lsprofile_something()
+        # if this worked we want 
+        # LSProf output for <built in function unicode> (['hello'], {'errors': 'replace'})
+        #    CallCount    Recursive    Total(ms)   Inline(ms) module:lineno(function)
+        # (the lsprof header)
+        # ... an arbitrary number of lines
+        # and the function call which is time.sleep.
+        #           1        0            ???         ???       ???(sleep) 
+        # and then repeated but with 'world', rather than 'hello'.
+        # this should appear in the output stream of our test result.
+        self.assertContainsRe(result_stream.getvalue(), 
+            r"LSProf output for <type 'unicode'>\(\('hello',\), {'errors': 'replace'}\)\n"
+            r" *CallCount *Recursive *Total\(ms\) *Inline\(ms\) *module:lineno\(function\)\n"
+            r" +1 +0 +0\.\d+ +0\.\d+ +<method 'disable' of '_lsprof\.Profiler' objects>\n"
+            r"LSProf output for <type 'unicode'>\(\('world',\), {'errors': 'replace'}\)\n"
+            r" *CallCount *Recursive *Total\(ms\) *Inline\(ms\) *module:lineno\(function\)\n"
+            r" +1 +0 +0\.\d+ +0\.\d+ +<method 'disable' of '_lsprof\.Profiler' objects>\n"
+            )
+
 
 class TestRunner(TestCase):
 
@@ -571,6 +618,26 @@
             output_stream.getvalue(),
             "[1-9][0-9]ms/   [1-9][0-9]ms\n$")
         
+    def test__gather_lsprof_in_benchmarks(self):
+        """When _gather_lsprof_in_benchmarks is on, accumulate profile data.
+        
+        Each self.time() call is individually and separately profiled.
+        """
+        try:
+            import bzrlib.lsprof
+        except ImportError:
+            raise TestSkipped("lsprof not installed.")
+        # overrides the class member with an instance member so no cleanup 
+        # needed.
+        self._gather_lsprof_in_benchmarks = True
+        self.time(time.sleep, 0.000)
+        self.time(time.sleep, 0.003)
+        self.assertEqual(2, len(self._benchcalls))
+        self.assertEqual((time.sleep, (0.000,), {}), self._benchcalls[0][0])
+        self.assertEqual((time.sleep, (0.003,), {}), self._benchcalls[1][0])
+        self.assertIsInstance(self._benchcalls[0][1], bzrlib.lsprof.Stats)
+        self.assertIsInstance(self._benchcalls[1][1], bzrlib.lsprof.Stats)
+
 
 class TestExtraAssertions(TestCase):
     """Tests for new test assertions in bzrlib test suite"""

